"""
This module contains classes that represent configuration files. Each class
contains information and methods to manipulate specific config file except
the parent (File) class that is supposed to operate on general config file.

Basic operations on config files defined in this module:
    * create
        creates content of config file usually based on template.
        Note, that some child classes may also update content of config
        file if it already existed and backup original file.
    * set
        modify values of config files, add keys or sections if necessary
    * save
        save modified content to config file
    * clean
        remove config file; note that some child classes may also restore
        original config file if backup exists.
"""
import os
from configparser import ConfigParser
from pathlib import Path
from shutil import copy2
from traceback import format_exc
from typing import Union
import json

from SCAutolib import logger, TEMPLATES_DIR, LIB_BACKUP, LIB_DUMP_CONFS, run
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.isDistro import isDistro


class File:
    """
    This class defines an interface for generic operations on config files

    * create: create content of config file based on template file
    * set:    modify content of config files, add keys or sections if necessary
    * save:   save config file
    * clean:  remove config file

    .. note:: Set method operates **only** on:

     * files compatible with ConfigParser (i.e. files containing sections)
     * simple config files without sections.

     Other formats of config files are not supported.
    """
    _conf_file = None
    _template = None
    _default_parser = None
    _simple_content = None

    def __init__(self, filepath: Union[str, Path], template: Path = None):
        """
        Init of File

        :param filepath: Path of config file
        :type filepath: str or pathlib.Path
        :param template: Path of template file
        :type template: str
        """
        self._conf_file = Path(filepath)
        if template is not None:
            self._template = template

    @property
    def path(self):
        return self._conf_file

    def create(self):
        """
        Populate internal parser object with content based on template.
        """
        if self._conf_file.exists():
            logger.warning(f"Create error: {self._conf_file} already exists.")
            raise FileExistsError(f'{self._conf_file} already exists')
        else:
            self._default_parser = ConfigParser()
            self._default_parser.optionxform = str
            if self._template is None:
                raise FileNotFoundError("Template file was not provided.")
            with self._template.open() as t:
                self._default_parser.read_file(t)

    def remove(self):
        """
        Removes the file if it exists.
        """
        if self._conf_file.exists():
            self._conf_file.unlink()
            logger.debug(
                f"Removed file {self._conf_file}."
            )

    def exists(self):
        """
        Checks if a file exists. Returns boolean.
        """
        return self._conf_file.exists()

    def set(self, key: str, value: Union[int, str, bool], section: str = None,
            separator: str = "="):
        """
        Modify value in config file. Modification is made through the
        ConfigParser object if it is defined. If not, then key value pair
        would be written to the file through normal :code:`write()` method with
        composed string in the following form :code:`<key><separator><value>`

        .. note::
            spaces around key has to be specified as a part of the
            :code:`separator` parameter.

        :param key: value for this key will be updated
        :type key: str
        :param value: new value to be stored in [section][key] of config file
        :type value: int or str or bool
        :param section: section of config file that will be modified
        :type section: str
        :param separator: Character to be used as a separator between key and
            value in files that are not supported by ConfigParser object.
        :type separator: str

        """
        if section is None:
            # simple config files without sections
            if self._simple_content is None:
                with self._conf_file.open() as config:
                    self._simple_content = config.readlines()
            modified = False
            new_content = []
            for line in self._simple_content:
                # skip comments and empty lines
                if line.strip().startswith("#") or len(line.strip()) == 0:
                    new_content.append(line)
                    continue
                try:
                    conf_key, conf_val = line.split(separator, 1)
                except ValueError:
                    raise ValueError(f"unexpected format of line: {line}")
                if conf_key.strip() == key:
                    new_content.append(line.replace(conf_val, value + '\n'))
                    modified = True
                else:
                    new_content.append(line)
            if not modified:
                new_content.append(f"\n{key}{separator}{value}")
            self._simple_content = new_content
        else:
            # configparser compatible config files (with sections)
            if self._default_parser is None:
                self._default_parser = ConfigParser()
                self._default_parser.optionxform = str
                with self._conf_file.open() as config:
                    self._default_parser.read_file(config)
            if not self._default_parser.has_section(section):
                logger.warning(f"Section {section} not present.")
                logger.info(f"Adding section {section}.")
                self._default_parser.add_section(section)

            previous = self._default_parser.get(section, key,
                                                fallback="Not set")

            self._default_parser.set(section, key, value)
            logger.info("Value is changed.")
            logger.debug(f"Old value in section [{section}] {key}={previous}")
            logger.debug(f"New value in section [{section}] {key}={value}")

    def get(self, key, section: str = None, separator: str = "="):
        """
        Method processes and returns the value of the key in section. If the
        section is not provided (section=None), then file would be parsed line
        by line splitting the line on separator. First match wins and is
        returned.

        If section is provided and the file can be parsed by the
        :code:`ConfigParser`, then this object would be used to look for the
        key.

        :param key: required key
        :param section: section where the key should be found
        :param separator: applicable only for non-configparser file. Separator
            that would be used to so split a line from the file. By default
            separator is '='
        :raise SCAutolib.SCAutolibException: if the key is not found the
            non-ConfigParser file
        :raise configparser.NoSectionError: if the section is not found in
            ConfigParser-supported file
        :raise KeyError: if the key is not present in ConfigParser-supported
            file
        :return: value of the key in section (if set)
        """
        if section is None:
            # simple config files without sections
            if self._simple_content is None:
                with self._conf_file.open() as config:
                    self._simple_content = config.readlines()
            for line in self._simple_content:
                if line.strip().startswith("#") or line.strip() == "":
                    continue
                key_from_file, value = line.split(separator, maxsplit=1)
                if key_from_file == key:
                    return value.strip()

            raise SCAutolibException(f"Key '{key}' doesn't present in the "
                                     f"file {self._conf_file}")
        elif self._default_parser is None:
            self._default_parser = ConfigParser()
            self._default_parser.optionxform = str
            with self._conf_file.open() as config:
                self._default_parser.read_file(config)
        return self._default_parser[section][key]

    def save(self):
        """
        Save content of config file stored in parser object to config file.
        """
        if self._simple_content is None:
            with self._conf_file.open("w") as config:
                self._default_parser.write(config)
        else:
            with self._conf_file.open("w") as config:
                config.writelines(self._simple_content)

    def clean(self):
        """
        Removes config file
        """
        try:
            self._conf_file.unlink()
            logger.info(f"Removing {self._conf_file}.")
        except FileNotFoundError:
            logger.info(f"{self._conf_file} does not exist. Nothing to do.")

    def backup(self, name: str = None):
        """
        Save original file to the backup directory with given name. If name is
        None, default name is :code:`<filename>.<extension>.backup`

        :param name: custom file name to be set for the file
        :type name: str
        :return: path where the file is stored
        """

        new_path = LIB_BACKUP.joinpath(
            f"{name if name else self._conf_file.name}.backup")
        copy2(self._conf_file, new_path)
        logger.debug(f"File {self._conf_file} is stored to {new_path}")
        self._backup = {"original": str(self._conf_file),
                        "backup": str(new_path)}
        return new_path

    def restore(self, name: str = None):
        """
        Copies backup file to original file location.
        """
        original_path = LIB_BACKUP.joinpath(
            f"{name if name else self._conf_file.name}.backup")

        if original_path.exists():
            with self._conf_file.open("w") as config, \
                    original_path.open() as backup:
                config.write(backup.read())
            original_path.unlink()
            logger.debug(
                f"File {self._conf_file} is restored to {original_path}"
            )
        else:
            logger.debug(
                f"File {self._conf_file} was not backed up. Nothing to do."
            )


class SSSDConf(File):
    """
    This class contains information and methods to create and modify
    /etc/sssd/sssd.conf file.

    It is implemented as singleton, which allows to use class object
    :code:`_default_parser` as representation of content of sssd.conf file
    during runtime.

    Intended use is to create/update and save config file in first runtime
    and load content of config file to internal parser object in following
    runtimes.
    """
    __instance = None
    _conf_file = Path("/etc/sssd/sssd.conf")
    _backup_original = None
    _backup_default = LIB_BACKUP.joinpath('default-sssd.conf')
    _backup_current_cont = None
    _before_last_change_cont = None
    _changed = False

    dump_file: Path = LIB_DUMP_CONFS.joinpath("SSSDConf.json")

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = super(SSSDConf, cls).__new__(cls)
            cls.__instance.__initialized = False
        return cls.__instance

    def __init__(self):
        if self.__initialized:
            return
        self.__initialized = True

        if isDistro(['rhel', 'centos'], version='<=9') \
                or isDistro(['fedora'], version='<39'):
            self._template = TEMPLATES_DIR.joinpath("sssd.conf-8or9")
        else:
            self._template = TEMPLATES_DIR.joinpath("sssd.conf-10")

        # _default_parser object stores default content of config file
        self._default_parser = ConfigParser()
        # avoid problems with inserting some 'specific' values
        self._default_parser.optionxform = str

        if self._backup_default.exists():
            with self._backup_default.open() as config:
                self._default_parser.read_file(config)

        # _changes parser object reflects modifications imposed by set method
        self._changes = ConfigParser()
        self._changes.optionxform = str

        if self.dump_file.exists():
            with self.dump_file.open("r") as f:
                cnt = json.load(f)
                self._backup_original = Path(cnt['_backup_original'])

    def __call__(self, key: str, value: Union[int, str, bool],
                 section: str = None):
        # We need to save the state of the current unchanged sssd.conf because
        # __call__ is called before __enter__ in
        # with SSSDConf(key, value, section):
        with self._conf_file.open() as config:
            self._before_last_change_cont = config.read()

        self.set(key, value, section)
        self.save()
        run("systemctl restart sssd", sleep=10)
        return self

    def __enter__(self):
        # Check if we changed the file or not and save version before context
        # manager was called
        if self._before_last_change_cont:
            self._backup_current_cont = self._before_last_change_cont
        else:
            with self._conf_file.open() as config:
                self._backup_current_cont = config.read()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._changed:
            # Restore sssd.conf to the version before context manager was
            # called
            with self._conf_file.open("w") as config:
                config.write(self._backup_current_cont)
            self._backup_current_cont = None
            self._before_last_change_cont = None
            self._changed = False
        if exc_type is not None:
            logger.error("Exception in virtual smart card context")
            logger.error(format_exc())
        run("systemctl restart sssd", sleep=10)

    def create(self):
        """
        Populate internal parser object with content from existing config file
        and update it with values from config template. Back up original files.
        """
        try:
            with self._conf_file.open() as config:
                self._default_parser.read_file(config)
            logger.info(f"{self._conf_file} file exists, loading values")
            self._backup_original = self.backup("sssd-conf-original")

        except FileNotFoundError:
            logger.warning(f"{self._conf_file} not present")
            logger.warning("Creating sssd.conf based on the template")

        with self._template.open() as template:
            logger.info(f"Updating {self._conf_file} with values from the "
                        f"template")
            self._default_parser.read_file(template)

        with self._backup_default.open("w") as bdefault:
            self._default_parser.write(bdefault)

        with self.dump_file.open("w") as f:
            json.dump({
                "_backup_original": str(self._backup_original)
            }, f)

    def set(self, key: str, value: Union[int, str, bool], section: str = None):
        """
        Modify or add content of config file represented by ConfigParser object

        If a value is set outside of a context manager, it is the user's
        responsibility to revert it.

        :param key: key from section of config file to be updated
        :type key: str
        :param value: new value to be stored in [section][key] of config file
        :type value: int or bool or str
        :param section: section of config file to be created/updated
        :type section: str
        """
        if len(self._changes.sections()) == 0:
            with self._conf_file.open() as config:
                self._changes.read_file(config)

        if not self._changes.has_section(section):
            logger.warning(f"Section {section} not present.")
            logger.info(f"Adding section {section}.")
            self._changes.add_section(section)

        previous = self._changes.get(section, key, fallback="Not set")
        if previous == value:
            logger.info(f"A key '{key}' in section '{section}' is already set "
                        f"to {value}. No changes in SSSD are required.")
            return

        self._changes.set(section, key, value)
        self._changed = True

        logger.info(f"Value is changed in section {self._changes[section]}")
        logger.debug(f"Old value in section [{section}] {key}={previous}")
        logger.debug(f"New value in section [{section}] {key}={value}")

    def save(self):
        """
        Save content of config file stored in parser object to config file.

        .. note: SSSD service restart is caller's responsibility.
        """
        with self._conf_file.open("w") as config:
            if len(self._changes.sections()) == 0:
                # after create; _changes is empty; content is in _default_parser
                self._default_parser.write(config)
            else:
                # after set; _changes reflects current content
                self._changes.write(config)
                # re-initialization because I did not find other simple way
                # to empty parser object
                self._changes = ConfigParser()
                self._changes.optionxform = str
        os.chmod(self._conf_file, 0o600)

    def restore(self):
        """
        Removes sssd.conf file in case it was created by this package or
        restore original sssd.conf in case the file was modified.

        .. note: SSSD service restart is caller's responsibility.
        """

        if self._backup_original and self._backup_original.exists():
            with self._backup_original.open() as original, \
                    self._conf_file.open("w") as config:
                config.write(original.read())
            self._backup_original.unlink()
        else:
            self.clean()

        if self._backup_default.exists():
            self._backup_default.unlink()

        if self.dump_file.exists():
            self.dump_file.unlink()
            logger.debug(f"Removed {self.dump_file} dump file")

        logger.info("Restored sssd.conf to the original version")
        self._changed = False

    def update_default_content(self):
        """
        Populate internal parser object with content from current config file.
        """
        self._default_parser = ConfigParser()
        self._default_parser.optionxform = str
        with self._conf_file.open() as config:
            self._default_parser.read_file(config)
        logger.info(f"Backing up {self._conf_file} as {self._backup_default}")
        copy2(self._conf_file, self._backup_default)

    def check_backups(self):
        """
        Raises an exception if internal backup files already exists
        """
        backup_files = (self._backup_default, self._backup_original)
        for file in backup_files:
            if file.exists():
                logger.error(f"Backup of {file} already exists")
                logger.error("This suggest that create method was already "
                             "executed. Create method should not be executed "
                             "multiple times")
                raise FileExistsError(f'{file} file exists')


class SoftHSM2Conf(File):
    """
    This class provide information and methods to create and modify
    softhsm2.conf file.
    """
    _template = Path(TEMPLATES_DIR, "softhsm2.conf")
    _conf_file = None
    _content = None
    _card_dir = None

    def __init__(self, filepath: Union[str, Path], card_dir: Union[str, Path]):
        """
        Init of SoftHSM2Conf

        :param filepath: path where config file should be saved
        :type filepath: str
        :param card_dir: parameter to be updated in config file
        :type card_dir: str
        """
        self._conf_file = filepath if isinstance(filepath, Path) else \
            Path(filepath)
        self._card_dir = card_dir if isinstance(card_dir, Path) else \
            Path(card_dir)

    def create(self):
        """
        Populate internal file object with content based on template.
        """
        with self._template.open('r') as template:
            self._content = template.read().format(card_dir=self._card_dir)

        logger.info(f"Creating content of {self._conf_file} "
                    f"based on {self._template}")

    def set(self, *args):
        """
        :raise NotImplementedError: if this method is called on SoftHSM2Conf.
        """
        # parent class set method does not work as softHSM2 conf does not have
        # sections. Method do modify softHSM2 conf is not implemented
        logger.warning("softhsm2.conf does not contain sections.")
        raise NotImplementedError("softHSM2conf.set method not implemented")

    def save(self):
        """
        Save content stored in internal file object to config file.
        """
        with self._conf_file.open("w") as config:
            config.write(self._content)
        logger.debug(f"Config file {self._conf_file} is created")


class OpensslCnf(File):
    """
    This class provides information and methods to create and modify
    openssl cnf files.
    """
    _template = None
    _conf_file = None
    _content = None
    _old_string = None
    _new_string = None

    # openssl cnf content depends substantially on its purpose and separate
    # templates are needed for specific config files types. mapping:
    types = {
        "CA": {"template": Path(TEMPLATES_DIR, 'ca.cnf'),
               "replace": ["{ROOT_DIR}"]},
        "user": {"template": Path(TEMPLATES_DIR, 'user.cnf'),
                 "replace": ["{user}", "{cn}"]}
    }

    def __init__(self, filepath: Union[str, Path], conf_type: str,
                 replace: Union[str, list]):
        """
        Init of opensslCNF

        :param filepath: Path of config file
        :type filepath: str or pathlib.Path
        :param conf_type: Identifier of cnf file
        :type conf_type: basestring
        :param replace: list of strings that will replace specific strings from
                        template
        :type replace: list
        """
        self._conf_file = Path(filepath)
        self._template = Path(self.types[conf_type]["template"])
        self._old_strings = self.types[conf_type]["replace"]
        if isinstance(replace, str):
            replace = [replace]
        self._new_strings = replace

    def create(self):
        """
        Populate internal file object with content based on template
        and update specific strings
        """
        with self._template.open('r') as template:
            self._content = template.read()
        for old, new in zip(self._old_strings, self._new_strings):
            self._content = self._content.replace(old, new)

    def save(self):
        """
        Save content stored in internal file object to config file.
        """
        with self._conf_file.open("w") as config:
            if self._default_parser is None:
                config.write(self._content)
            else:
                # in case set method was used
                self._default_parser.write(config)



"""
This module defines the User and IPAUser classes which can be used
to represent system and IPA users.

The classes contain the usual properties that defines a user, like username,
password, etc.

The classes implement add_user and delete_user methods which can be used to
create or remove a specified user in the system or in the specified IPA server.
"""
import json
import pwd
import python_freeipa
from pathlib import Path, PosixPath

from SCAutolib import run, logger, LIB_DUMP_USERS
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.models.CA import IPAServerCA
from SCAutolib.enums import UserType


class User:
    """
    User represents general system user.
    """
    username: str = None
    password: str = None
    dump_file: Path = None
    user_type: str = None

    def __init__(self, username, password):
        self.username = username
        self.password = password
        self.user_type = UserType.local
        self.dump_file = LIB_DUMP_USERS.joinpath(f"{self.username}.json")

    def to_dict(self):
        # Retype patlib.Path object to str
        d = {k: str(v) if type(v) in (PosixPath, Path) else v
             for k, v in self.__dict__.items()}
        return d

    @staticmethod
    def load(json_file, **kwargs):
        """
        Load values from user's JSON file to corresponding user object.

        :param json_file: path to JSON file to read from
        :type json_file: pathlib.Path
        :param kwargs: dictionary of additional values needed to initialise the
            object
        :type kwargs: dict
        :return: user object
        :rtype: SCAutolib.models.user.User or SCAutolib.models.user.IPAUser
        """
        with json_file.open("r") as f:
            cnt = json.load(f)

        if cnt["user_type"] == UserType.local:
            user = User(username=cnt["username"],
                        password=cnt["password"])

        elif cnt["user_type"] == UserType.ipa:
            if "ipa_server" not in kwargs:
                raise SCAutolibException("IPA Server object was not provided. "
                                         "Can't load IPA user.")

            user = IPAUser(ipa_server=kwargs["ipa_server"],
                           username=cnt["username"],
                           password=cnt["password"])

        else:
            raise SCAutolibException(f"Unknown user type: {cnt['user_type']}")

        logger.debug(f"User {user.__class__} is loaded: {user.__dict__}")

        return user

    def add_user(self):
        """
        Add user to the local system with `useradd` bash command and set
        password for created user.
        """
        try:
            pwd.getpwnam(self.username)
            msg = f"User {self.username} already exists on this " \
                  f"machine. Username should be unique to avoid " \
                  f"future problems with collisions"
            logger.critical(msg)
            raise SCAutolibException(msg)
        except KeyError:
            logger.debug(f"Creating new user {self.username}")
            cmd = ['useradd', '-m', self.username]
            run(cmd, check=True)
            cmd = ["passwd", self.username, "--stdin"]
            run(cmd, input=self.password)
            logger.info(f"User {self.username} was added to the system")

    def delete_user(self):
        """
        Deletes the user
        """
        try:
            pwd.getpwnam(self.username)
            logger.info(f"Deleting the user {self.username}")
            run(['userdel', '-f', self.username], check=True)
        except KeyError:
            logger.info(f"User {self.username} is not present on the system")

        if self.dump_file.exists():
            self.dump_file.unlink()
            logger.debug(f"Removed {self.dump_file} dump file")


class IPAUser(User):
    """
    This class represents an IPA user.
    """
    default_password = "redhat"

    def __init__(self, ipa_server: IPAServerCA, *args, **kwargs):
        """
        Class for IPA user. IPA client should be configured first before
        creating an IPA user through this class.

        :param ipa_server: IPAServerCA object which provides the ipa hostname
        :type ipa_server: IPAServerCA
        :param username: Username for the system user
        :type username: str
        :param password: Password for the system user
        :type password: str
        :param key: Key to be associated with the user
        :type key: Path
        :param cert: Certificate to be associated with the user.
        :type cert: Path
        """

        super().__init__(*args, **kwargs)
        self.user_type = UserType.ipa
        self._meta_client = ipa_server.meta_client
        self._ipa_hostname = ipa_server.ipa_server_hostname

    def to_dict(self):
        d = super().to_dict()
        d.pop("_meta_client")
        d.pop("_ipa_hostname")
        return d

    def add_user(self):
        """
        Adds IPA user to IPA server.
        """
        try:
            r = self._meta_client.user_add(self.username, self.username,
                                           self.username, self.username,
                                           o_userpassword=self.default_password)
            logger.debug(r)

            # To avoid forcing IPA server to change the password on first login
            # we changing it through the client
            client = python_freeipa.client.Client(self._ipa_hostname,
                                                  verify_ssl=False)
            client.change_password(self.username, self.password,
                                   self.default_password)
            logger.info(f"User {self.username} is added to the IPA server")
        except python_freeipa.exceptions.DuplicateEntry:
            msg = f"User {self.username} already exists on the " \
                  f"IPA server. Username should be unique to avoid " \
                  f"future problems with collisions"
            logger.critical(msg)
            raise SCAutolibException(msg)

    def delete_user(self):
        """
        Deletes the user
        """
        try:
            r = self._meta_client.user_del(self.username)["result"]
            logger.info(f"User {self.username} is removed from the IPA server")
            logger.debug(r)
        except python_freeipa.exceptions.NotFound:
            pass



from shutil import copy

from SCAutolib import utils, LIB_DUMP_USERS


def test_load_user(local_user, tmp_path):
    utils.dump_to_json(local_user)
    copy(local_user.dump_file, LIB_DUMP_USERS.joinpath(
        f"{local_user.username}.json"))

    user = utils.load_user(local_user.username)
    assert user



"""
This module contains unit tests for file.py and related functions
"""
import filecmp
import shutil
from pathlib import Path

import pytest

from SCAutolib.models.file import SSSDConf
from conftest import FILES_DIR
from test_sssd_conf import compare_conf_files


def test_create_fail(file_test_prepare):
    """
    Exception is raised if conf file already exists
    """
    file_test = file_test_prepare
    with file_test._conf_file.open('w') as config:
        config.write('Test config file')
    with pytest.raises(FileExistsError, match=f'{file_test._conf_file} '
                                              f'already exists'):
        file_test.create()


def test_create(file_test_prepare):
    """
    Content of config file is created based on template if the file is not
    present in the system.
    """
    file_test = file_test_prepare
    file_test._template = SSSDConf._template
    file_test.create()
    # save content to file so it can easily be compared
    file_test.save()
    compare_conf_files(SSSDConf._template, file_test._conf_file)


def test_set_add_value(file_test_prepare):
    """
    Tests that new value is added to existing section
    """
    file_test = file_test_prepare
    file_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")
    file_test.set("newkey", "newvalue", "testsection")
    assert file_test._default_parser["testsection"]["newkey"] == "newvalue"


def test_set_add_section(file_test_prepare):
    """
    Tests that section is added with key and value if the section wasn't present
    """
    file_test = file_test_prepare
    file_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")
    file_test.set("newkey", "newvalue", "newsection")
    assert file_test._default_parser["newsection"]["newkey"] == "newvalue"


def test_set_overwrite_value(file_test_prepare):
    """
    Tests that value in is overwritten if section and key already existed
    """
    file_test = file_test_prepare
    file_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")
    file_test.set("testkey", "overwrite", "testsection")
    assert file_test._default_parser["testsection"]["testkey"] == "overwrite"


def test_set_simple_content(file_test_prepare):
    """
    Tests that value in is overwritten if section and key already existed
    """
    file_test = file_test_prepare
    test_start = Path(FILES_DIR, "conf_file_start")
    shutil.copy2(test_start, file_test._conf_file)

    file_test.set("prompt", "newprompt")
    file_test.set("OU", "testvalue")
    file_test.set("newkey", "newvalue")
    file_test.save()

    test_result = Path(FILES_DIR, "conf_file_end")
    assert filecmp.cmp(test_result, file_test._conf_file, shallow=False)



"""
Implementation of CLI commands for SCAutolib.
"""

import click
from pathlib import Path
from sys import exit

from collections import OrderedDict

from SCAutolib import logger, exceptions, schema_user
from SCAutolib.controller import Controller
from SCAutolib.enums import ReturnCode


def check_conf_path(conf):
    return click.Path(exists=True, resolve_path=True)(conf)


# In Help output, force the subcommand list to match the order
# listed in this file.   Solution was found here:
# https://github.com/pallets/click/issues/513#issuecomment-301046782
class NaturalOrderGroup(click.Group):
    """
    Command group trying to list subcommands in the order they were added.
    Example use::

    @click.group(cls=NaturalOrderGroup)

    If passing dict of commands from other sources, ensure they are of type
    OrderedDict and properly ordered, otherwise order of them will be random
    and newly added will come to the end.
    """
    def __init__(self, name=None, commands=None, **attrs):
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        click.Group.__init__(self, name=name,
                             commands=commands,
                             **attrs)

    def list_commands(self, ctx):
        """
        List command names as they are in commands dict.

        If the dict is OrderedDict, it will preserve the order commands
        were added.
        """
        return self.commands.keys()


@click.group(cls=NaturalOrderGroup)
@click.option("--conf", "-c",
              default="./conf.json",
              show_default=True,
              help="Path to JSON configuration file.")
@click.option('--force', "-f", is_flag=True, default=False, show_default=True,
              help="Force the command to overwrite configuration if it exists.")
@click.option("--verbose", "-v", default="DEBUG", show_default=True,
              type=click.Choice(
                  ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                  case_sensitive=False),
              help="Verbosity level.")
@click.pass_context
def cli(ctx, force, verbose, conf):
    logger.setLevel(verbose)
    ctx.ensure_object(dict)  # Create a dict to store the context
    ctx.obj["FORCE"] = force  # Store the force option in the context
    parsed_conf = None
    if ctx.invoked_subcommand != "gui":
        parsed_conf = check_conf_path(conf)
    ctx.obj["CONTROLLER"] = Controller(parsed_conf)


@cli.command()
@click.option("--gdm", "-g",
              required=False,
              default=False,
              is_flag=True,
              help="Install GDM package")
@click.option("--graphical",
              required=False,
              default=False,
              is_flag=True,
              help="Install dependencies for GUI testing module")
@click.option("--install-missing", "-i",
              required=False,
              default=False,
              is_flag=True,
              help="Install missing packages")
@click.pass_context
def prepare(ctx, gdm, install_missing, graphical):
    """Configure entire system for smart cards based on the config file."""
    ctx.obj["CONTROLLER"].prepare(
        ctx.obj["FORCE"],
        gdm,
        install_missing,
        graphical
    )
    exit(ReturnCode.SUCCESS.value)


@cli.command()
@click.option("--ca-type", "-t",
              required=False,
              default='all',
              type=click.Choice(['all', 'local', 'ipa'], case_sensitive=False),
              show_default=True,
              help="Type of the CA to be configured. If not set, all CA's "
                   "from the config file would be configured")
@click.pass_context
def setup_ca(ctx, ca_type):
    """
    Configure the CA's in the config file. If more than one CA is
    specified, specified CA type would be configured.
    """
    cnt = ctx.obj["CONTROLLER"]
    if ca_type == 'all':
        cnt.setup_local_ca(force=ctx.obj["FORCE"])
        cnt.setup_ipa_client(force=ctx.obj["FORCE"])
    elif ca_type == 'local':
        cnt.setup_local_ca(force=ctx.obj["FORCE"])
    elif ca_type == 'ipa':
        cnt.setup_ipa_client(force=ctx.obj["FORCE"])
    exit(ReturnCode.SUCCESS.value)


@cli.command()
@click.argument("name",
                required=True,
                default=None)
@click.option("--card-dir", "-d",
              required=False,
              default=None,
              help="Path to the directory where smart card should be created")
@click.option("--card-type", "-t",
              required=False,
              default="virtual",
              type=click.Choice(
                  ["virtual", "real", "removinator"], case_sensitive=False),
              show_default=True,
              help="Type of the smart card to be created")
@click.option("--passwd", "-p",
              required=False,
              default=None,
              show_default=True,
              help="Password for the user")
@click.option("--pin", "-P",
              required=False,
              default=None,
              show_default=True,
              help="PIN for the smart card")
@click.option("--user-type", "-T",
              required=False,
              default="local",
              type=click.Choice(["local", "ipa"], case_sensitive=False),
              show_default=True,
              help="Type of the user to be created")
@click.pass_context
def setup_user(ctx, name, card_dir, card_type, passwd, pin, user_type):
    """Configure user with smart cards (if set) based on the config file."""
    cnt = ctx.obj["CONTROLLER"]
    logger.info(f"Start setup user {name}")
    try:
        user_dict = cnt.get_user_dict(name)
    except exceptions.SCAutolibMissingUserConfig:
        logger.warning(f"User {name} not found in config file, "
                       f"trying to create a new one")
        if not all([card_dir, card_type, passwd, pin, user_type]):
            logger.error("Not all required arguments are set")
            logger.error("Required arguments: --card-dir, --pin, --password")
            exit(ReturnCode.ERROR.value)
        user_dict = schema_user.validate(
            {"name": name,
             "card_dir": Path(card_dir),
             "card_type": card_type,
             "passwd": passwd,
             "pin": pin,
             "local": user_type == "local"})
        logger.debug(f"User dict: {user_dict}")

    try:
        cnt.init_ca(user_dict["local"])
    except exceptions.SCAutolibMissingCA:
        logger.error("CA is not configured on the system")
        exit(ReturnCode.MISSING_CA.value)

    try:
        user = cnt.setup_user(user_dict, ctx.obj["FORCE"])
        cnt.enroll_card(user, ctx.obj["FORCE"])
    except exceptions.SCAutolibException:
        logger.error("Something went wrong")
        exit(ReturnCode.FAILURE.value)
    exit(ReturnCode.SUCCESS.value)


@cli.command()
@click.pass_context
def cleanup(ctx):
    """
    Cleanup all the configurations and system changes done by the prepare
    command.
    """
    ctx.obj["CONTROLLER"].cleanup()
    exit(ReturnCode.SUCCESS.value)


@cli.group(cls=NaturalOrderGroup, chain=True)
@click.option("--install-missing", "-i",
              required=False,
              default=False,
              is_flag=True,
              help="Install missing packages")
@click.pass_context
def gui(ctx, install_missing):
    """ Run GUI Test commands """
    pass


@gui.command()
def init():
    """ Initialize GUI for testing """
    return "init"


@gui.command()
@click.option("--no",
              required=False,
              default=False,
              is_flag=True,
              help="Reverse the action")
@click.argument("name")
def assert_text(name, no):
    """ Check if a word is found on the screen """
    if no:
        return f"assert_no_text:{name}"
    return f"assert_text:{name}"


@gui.command()
@click.argument("name")
def click_on(name):
    """ Click on object containing word """
    return f"click_on:{name}"


@gui.command()
@click.option("--no",
              required=False,
              default=False,
              is_flag=True,
              help="Reverse the action")
def check_home_screen(no):
    """ Check if screen appears to be the home screen """
    if no:
        return "check_no_home_screen"
    return "check_home_screen"


@gui.command()
@click.argument("keys")
def kb_send(keys):
    """ Send key(s) to keyboard """
    return f"kb_send:{keys}"


@gui.command()
@click.argument("keys")
def kb_write(keys):
    """ Send string to keyboard """
    return f"kb_write:{keys}"


@gui.command()
def done():
    """ cleanup after testing """
    return "done"


@gui.result_callback()
@click.pass_context
def run_all(ctx, actions, install_missing):
    """ Run all cli actions in order """
    ctx.obj["CONTROLLER"].setup_graphical(install_missing, True)

    from SCAutolib.models.gui import GUI
    gui = GUI(from_cli=True)
    for action in actions:
        if "init" in action:
            gui.__enter__()
        if "assert_text" in action:
            assert_text = action.split(":", 1)[1]
            gui.assert_text(assert_text)
        if "assert_no_text" in action:
            assert_text = action.split(":", 1)[1]
            gui.assert_no_text(assert_text)
        if "click_on" in action:
            click_on = action.split(":", 1)[1]
            gui.click_on(click_on)
        if "check_home_screen" in action:
            gui.check_home_screen()
        if "check_no_home_screen" in action:
            gui.check_home_screen(False)
        if "kb_send" in action:
            params = action.split(":", 1)[1].split()[0]
            gui.kb_send(params)
        if "kb_write" in action:
            params = action.split(":", 1)[1].split()[0]
            gui.kb_write(params)
            gui.kb_send('enter')
        if "done" in action:
            gui.__exit__(None, None, None)
            ctx.obj["CONTROLLER"].cleanup()


if __name__ == "__main__":
    cli()



"""
This module provides methods allowing to configure system for smart-card
authentication implemented as context manager. It calls authselect tool (see man
authselect(8)) and applies sssd profile with selected 'Authselect profile
features' (see man authselect-migration(7)).
"""
from os.path import exists

from traceback import format_exc

from SCAutolib import LIB_BACKUP
from SCAutolib import logger
from SCAutolib.utils import run


class Authselect:
    backup_name = LIB_BACKUP.joinpath("SCAutolib_authselect_backup")

    def __init__(self, required=False, lock_on_removal=False,
                 mk_homedir=False, sudo=False):
        """
        Constructor for Authselect class. Only 'with-smartcard' feature of sssd
        profile is set by default and --force option is used if the sssd profile
        is modified. Previous configuration is backed up and restored on exiting
        context manager.

        :param required: specifies with-smartcard-required option
        :type required: bool
        :param lock_on_removal: specifies with-smartcard-lock-on-removal option
        :type lock_on_removal: bool
        :param mk_homedir: specifies with-mkhomedir option
        :type mk_homedir: bool
        """

        self._options = ["with-smartcard"]
        if required:
            self._options.append("with-smartcard-required")
        if lock_on_removal:
            self._options.append("with-smartcard-lock-on-removal")
        if mk_homedir:
            self._options.append("with-mkhomedir")
        if sudo:
            self._options.append("with-sudo")

    def _set(self):
        """
        Set authselect with SSSD profile and set selected Authselect profile
        features. Features are passed into the constructor.
        """

        # compose and run Authselect command
        cmd = ["authselect", "select", "sssd", *self._options,
               "--backup", self.backup_name, "--force"]
        run(cmd)

        # get modified setup and log it
        logger.debug("Current Authselect setting is:")
        run(["authselect", "current"], return_code=[0, 2])

        logger.debug(f"Original Authselect configuration was backed up with "
                     f"authselect to default location as : "
                     f"{str(self.backup_name)}")
        logger.debug("Default location is: /var/lib/authselect/backups/")

    def _restore(self):
        """
        Restore the previous configuration of authselect.
        """
        if exists(f"/var/lib/authselect/backups/{self.backup_name}"):
            cmd = ["authselect", "backup-restore", self.backup_name, "--debug"]
            run(cmd)
            logger.debug("Authselect configuration is restored to:")
            run(["authselect", "current"], return_code=[0, 2])
        else:
            # as _set and _restore should be used in context manager defined in
            # this class, it should not happen that backup does not exist except
            # something failed, or it's misused
            raise FileNotFoundError("Backup file not found. _restore method was"
                                    "probably called in unexpected manner.")

    def __enter__(self):
        self._set()
        return self

    def __exit__(self, ext_type, ext_value, ext_traceback):
        if ext_type is not None:
            logger.error("Exception in authselect context")
            logger.error(format_exc())
        self._restore()



import pytest
from pathlib import Path
from subprocess import check_output, run
from time import sleep

from SCAutolib.models.card import Card
from SCAutolib.models.file import SoftHSM2Conf
from SCAutolib.models.user import User
from SCAutolib.utils import dump_to_json


@pytest.fixture()
def gen_key_and_cert(local_ca_fixture, local_user):
    # FIXME this should be separate unit test - not just a helper function and
    #  it should really test that key and cert are created and no errors occur
    csr = Path(local_user.card_dir, f"{local_user.username}.csr")
    cert = Path(local_user.card_dir, f"{local_user.username}.cert")
    key = Path(local_user.card_dir, f"{local_user.username}.key")
    cmd = ["openssl", "req", "-new", "-nodes", "-newkey", "rsa:2048",
           "-keyout", key, "-out", csr, "-subj", f"/CN={local_user.username}"]
    check_output(cmd, encoding="utf-8")

    local_ca_fixture.request_cert(csr, username=local_user.username,
                                  cert_out=cert)
    return key, cert


@pytest.fixture()
def local_user_with_smart_card(local_user, gen_key_and_cert):
    # FIXME key and cert should be provided by test environment instead of
    #  creating them by fixture 'gen_key_and_cert'

    # FIXME softhsm conf should be provided and not created by this fixture
    #  while separate unit test should be created to test SoftHSM2Conf class
    hsm_conf = SoftHSM2Conf(Path(local_user.card_dir, "softhsm2.conf"),
                            local_user.card_dir)
    hsm_conf.create()
    hsm_conf.save()

    local_user.key, local_user.cert = gen_key_and_cert
    local_user.card.softhsm2_conf = hsm_conf.path
    yield local_user

    if local_user.card.service_location and \
            local_user.card.service_location.exists():
        local_user.card.service_location.unlink()


@pytest.mark.skip(reason="need to be fixed for compatibility with V3")
@pytest.mark.service_restart
def test_create_and_enroll(local_user_with_smart_card):
    # FIXME this is integration test and thus it should be moved to separate
    #  directory. SoftHSM token creation should be covered by separate test.
    sc = local_user_with_smart_card.card
    user = local_user_with_smart_card
    sc.create()

    sc.enroll()

    assert sc.uri, "Smart card URI is not set"

    # Test phase
    # FIXME service creation may be tested separately by unit test
    check_output(["systemctl", "start", sc._service_name], encoding="utf-8")
    sleep(3)
    try:
        out = check_output(["pkcs11-tool", "-O", "-l", "-p", f"{user.pin}"],
                           encoding="utf-8")
        for pattern in ["Private Key Object; RSA",
                        "Certificate Object; type = X.509 cert"]:
            assert pattern in out, "Probably some object is not correctly " \
                                   "uploaded to the virtual card"
    finally:
        # Cleanup makes sense only if the service is successfully started
        check_output(["systemctl", "stop", sc._service_name], encoding="utf-8")


@pytest.mark.skip(reason="need to be fixed for compatibility with V3")
@pytest.mark.service_restart
def test_context_manager(local_user_with_smart_card):
    # FIXME this is integration test. Integration tests or tests of context
    #  managers should be in separate directory. Objects needed for integration
    #  tests should be created centrally (i.e. once not with each test)
    sc = local_user_with_smart_card.card
    sc.create()
    sc.enroll()

    with sc as sc:
        sc.insert()
        out = check_output(["pkcs11-tool", "-O", "-l", "-p",
                            f"{local_user_with_smart_card.pin}"],
                           encoding="utf-8")
        for pattern in ["Private Key Object; RSA",
                        "Certificate Object; type = X.509 cert"]:
            assert pattern in out, "Probably some object is not correctly " \
                                   "uploaded to the virtual card"
    proc = run(["systemctl", "status", sc._service_name])
    assert proc.returncode == 3  # Service is not active


@pytest.mark.skip(reason="need to be fixed for compatibility with V3")
@pytest.mark.service_restart
def test_load_user_with_card(local_user_with_smart_card):
    # FIXME this is an integration test. necessary objects (user and card)
    #  should not be created within the test and instead they should be provided
    #  to the test
    local_user_with_smart_card.card.create()
    local_user_with_smart_card.card.enroll()

    dump_to_json(local_user_with_smart_card.card)
    dump_to_json(local_user_with_smart_card)

    user, card_file = User.load(local_user_with_smart_card.dump_file)
    card = Card.load(card_file, user=user)

    assert card.uri == local_user_with_smart_card.card.uri

    card.insert()
    card.remove()



import json
import pytest
from click.testing import CliRunner
from pwd import getpwnam
from subprocess import run

import SCAutolib.cli_commands as cli_cmd
from SCAutolib.cli_commands import ReturnCode
from SCAutolib.controller import Controller


@pytest.fixture(scope="module", autouse=True)
def runner():
    return CliRunner()


@pytest.mark.skip(reason="broken test")
@pytest.mark.parametrize("username", ["local-user", "not-existing-user"])
@pytest.mark.parametrize("ca", ["local", None])
@pytest.mark.service_restart
def test_cli_setup_user(dummy_config, runner, username, ca, tmpdir):
    """Test the CLI command setup-user."""
    # FIXME this is rather complex integration test. It's failing and its
    #  difficult to debug it as logging is unexpectedly weak. We should probably
    #  test that cli recognize parameters, however, testing of cli by executing
    #  real commands is arguable.
    if ca == "local":
        Controller(dummy_config).setup_local_ca()
    user_options = []
    if username == "not-existing-user":
        user_options = ["--pin", "123456", "--passwd", "user-passwd",
                        "--card-dir", tmpdir]
    result = runner.invoke(cli_cmd.cli, ["--conf", dummy_config,
                                         "setup-user", username,
                                         *user_options])
    print(result.stdout)
    try:
        if ca == "local":
            if username == "not-existing-user":
                assert f"User {username} not found in config file, trying to " \
                       f"create a new one" in result.output
            assert result.exit_code == ReturnCode.SUCCESS.value
            assert getpwnam(username)
        else:
            assert "CA is not configured on the system" in result.output
    finally:
        run(["userdel", "-r", username], check=False)


@pytest.mark.skip(reason="ipa server not available for tests")
@pytest.mark.ipa
@pytest.mark.parametrize("ca_type", ["all", "local", "ipa"])
def test_setup_ca(runner, dummy_config, ca_type, clean_ipa, ipa_config):
    """Test the CLI command setup-ca."""

    if ca_type in ["all", "ipa"]:
        with open(dummy_config, "r") as f:
            cfg = json.load(f)
        cfg["ca"]["ipa"]["ip_addr"] = ipa_config["ip"]
        cfg["ca"]["ipa"]["server_hostname"] = ipa_config["hostname"]
        cfg["ca"]["ipa"]["admin_passwd"] = ipa_config["admin_passwd"]
        cfg["ca"]["ipa"]["root_passwd"] = ipa_config["root_passwd"]
        cfg["ca"]["ipa"]["domain"] = ipa_config["domain"]
        cfg["ca"]["ipa"]["realm"] = ipa_config["domain"].upper()
        with open(dummy_config, "w") as f:
            json.dump(cfg, f)

    result = runner.invoke(cli_cmd.cli, ["--conf", dummy_config,
                                         "setup-ca", "--ca-type", ca_type])

    print(result.stdout)
    assert result.exit_code == ReturnCode.SUCCESS.value



"""
This module implements classes for communication with different types of cards
that we are using in the library. Those types are: virtual smart card, real
(physical) smart card in standard reader, cards in the removinator.
"""
import json
import re
import time
import shutil
from pathlib import Path
from traceback import format_exc

from SCAutolib import run, logger, TEMPLATES_DIR, LIB_DUMP_CARDS
from SCAutolib.exceptions import SCAutolibException
from SCAutolib.enums import CardType, UserType


class Card:
    """
    Interface for child classes. All child classes will rewrite common methods
    based on the type of the card.
    """
    uri: str = None
    dump_file: Path = None
    type: str = None
    _pattern: str = None

    def _set_uri(self):
        """
        Sets URI for given smart card. Uri is matched from ``p11tool`` command
        with regular expression. If URI is not matched, exception is raised.

        :raise: SCAutolibException
        """
        cmd = ["p11tool", "--list-token-urls"]
        out = run(cmd).stdout
        urls = re.findall(self._pattern, out)
        if len(urls) == 1:
            self.uri = urls[0]
            logger.info(f"Card URI is set to {self.uri}")
        elif len(urls) == 0:
            logger.warning("URI not set")
            raise SCAutolibException("URI matching expected pattern not found.")
        else:
            logger.warning("Multiple matching URIs found. URI not set")
            raise SCAutolibException("Multiple URIs match expected pattern.")

    def insert(self):
        """
        Insert the card.
        """
        ...

    def remove(self):
        """
        Remove the card.
        """
        ...

    def enroll(self):
        """
        Enroll the card (upload a certificate and a key on it)
        """
        ...

    @staticmethod
    def load(json_file):
        with json_file.open("r") as f:
            cnt = json.load(f)

        card = None
        if cnt["card_type"] == CardType.virtual:
            card = VirtualCard(cnt, softhsm2_conf=Path(cnt["softhsm"]))
#            card.uri = cnt["uri"]
        elif cnt["card_type"] == CardType.physical:
            card = PhysicalCard(cnt)
        else:
            raise SCAutolibException(
                f"Unknown card type: {cnt['card_type']}")
        return card


class VirtualCard(Card):
    """
    This class provides methods for operations on virtual smart card. Virtual
    smart card by itself is represented by the systemd service in the system.
    The card relates to some user, so providing the user is essential for
    correct functioning of methods for the virtual smart card.

    Card root directory has to be created before calling any method
    """

    _service_name: str = None
    _service_location: Path = None
    _softhsm2_conf: Path = None
    _nssdb: Path = None
    _template: Path = Path(TEMPLATES_DIR, "virt_cacard.service")
    _pattern = r"(pkcs11:model=PKCS%2315%20emulated;" \
               r"manufacturer=Common%20Access%20Card;serial=.*)"
    _inserted: bool = False

    name: str = None
    pin: str = None
    cardholder: str = None
    CN: str = None
    UID: str = None
    key: Path = None
    cert: Path = None
    card_dir: Path = None
    card_type: str = None
    ca_name: str = None
    slot: str = None
    user = None
    cnf = None

    def __init__(self, card_data, softhsm2_conf: Path = None,
                 card_dir: Path = None, key: Path = None, cert: Path = None):
        """
        Initialise virtual smart card. Constructor of the base class is also
        used.

        :param card_data: dict containing card details as pin, cardholder etc.
        :type card_data: dict
        :param softhsm2_conf: path to SoftHSM2 configuration file
        :type softhsm2_conf: pathlib.Path
        :param card_dir: path to card directory where card files will be saved
        :type card_dir: pathlib.Path
        :param key: path to key - if the key exist it will be used with the card
        :type key: pathlib.Path
        :param cert: path to certificate. If file exist it will be used with the
            card
        :type cert: pathlib.Path
        """
        self.name = card_data["name"]
        self.pin = card_data["pin"]
        self.cardholder = card_data["cardholder"]
        self.card_type = card_data["card_type"]
        self.CN = card_data["CN"]
        self.ca_name = card_data["ca_name"]
        self.card_dir = card_dir if card_dir is not None \
            else Path(card_data["card_dir"])
        if not self.card_dir.exists():
            raise FileNotFoundError("Card root directory doesn't exists")
        self.dump_file = LIB_DUMP_CARDS.joinpath(f"{self.name}.json")
        self.key = key \
            if key else self.card_dir.joinpath(f"key-{self.name}.pem")
        self.cert = cert \
            if cert else self.card_dir.joinpath(f"cert-{self.name}.pem")
        self._service_name = self.name
        self._service_location = Path(
            f"/etc/systemd/system/{self._service_name}.service")
        self._nssdb = self.card_dir.joinpath("db")
        self._softhsm2_conf = softhsm2_conf if softhsm2_conf \
            else Path(self.card_dir, "softhsm2.conf")

    def __call__(self, insert: bool):
        """
        Call method for virtual smart card. It would be used in the context
        manager.

        :param insert: True if the card should be inserted, False otherwise
        :type insert: bool
        """
        if insert:
            self.insert()
        return self.__enter__()

    def __enter__(self):
        """
        Start of context manager for virtual smart card. The card would be
        inserted if ``insert`` parameter in constructor is specified.

        :return: self
        """
        if not self._service_location.exists():
            raise FileNotFoundError("Service for virtual sc doesn't exists.")
        return self

    def __exit__(self, exp_type, exp_value, exp_traceback):
        """
        End of context manager for virtual smart card. If any exception was
        raised in the current context, it would be logged as an error.

        :param exp_type: Type of the exception
        :param exp_value: Value for the exception
        :param exp_traceback: Traceback of the exception
        """
        if exp_type is not None:
            logger.error("Exception in virtual smart card context")
            logger.error(format_exc())
        if self._inserted:
            self.remove()

    def to_dict(self):
        return {
            "name": self.name,
            "pin": self.pin,
            "cardholder": self.cardholder,
            "card_type": self.card_type,
            "CN": self.CN,
            "card_dir": str(self.card_dir),
            "key": str(self.key),
            "cert": str(self.cert),
            "uri": self.uri,
            "softhsm": str(self._softhsm2_conf),
            "ca_name": self.ca_name,
            "slot": self.slot
        }

    @property
    def softhsm2_conf(self):
        return self._softhsm2_conf

    @softhsm2_conf.setter
    def softhsm2_conf(self, conf: Path):
        if not conf.exists():
            raise FileNotFoundError(f"File {conf} doesn't exist")
        self._softhsm2_conf = conf

    @property
    def service_location(self):
        return self._service_location

    def insert(self):
        """
        Insert virtual smart card by starting the corresponding service.
        """
        cmd = ["systemctl", "start", self._service_name]
        out = run(cmd, check=True)
        time.sleep(2)  # to prevent error with fast restarting of the service
        logger.info(f"Smart card {self._service_name} is inserted")
        self._inserted = True
        return out

    def remove(self):
        """
        Remove the virtual card by stopping the service
        """
        cmd = ["systemctl", "stop", self._service_name]
        out = run(cmd, check=True)
        time.sleep(2)  # to prevent error with fast restarting of the service
        logger.info(f"Smart card {self._service_name} is removed")
        self._inserted = False
        return out

    def enroll(self):
        """
        Upload certificate and private key to the virtual smart card (upload to
        NSS database) with pkcs11-tool.
        """
        cmd = ["pkcs11-tool", "--module", "libsofthsm2.so", "--slot-index",
               '0', "-w", self.key, "-y", "privkey", "--label",
               "test_key", "-p", self.pin, "--set-id", "0",
               "-d", "0"]
        run(cmd, env={"SOFTHSM2_CONF": self._softhsm2_conf})
        logger.debug(
            f"User key {self.key} is added to virtual smart card")

        cmd = ['pkcs11-tool', '--module', 'libsofthsm2.so', '--slot-index', "0",
               '-w', self.cert, '-y', 'cert', '-p', self.pin,
               '--label', "test_cert", '--set-id', "0", '-d', "0"]
        run(cmd, env={"SOFTHSM2_CONF": self._softhsm2_conf})
        logger.debug(
            f"User certificate {self.cert} is added to virtual smart card")

        # To get URI of the card, the card has to be inserted
        # Virtual smart card can't be started without a cert and a key uploaded
        # to it, so URI can be set only after uploading the cert and a key
        with self:
            self.insert()
            self._set_uri()

    def create(self):
        """
        Creates SoftHSM2 token and systemd service for virtual smart card.
        Directory for NSS database is created in this method as separate DB is
        required for each virtual card.
        """

        if not self._softhsm2_conf.exists():
            raise FileNotFoundError("Can't proceed, SoftHSM2 conf not found.")

        self.card_dir.joinpath("tokens").mkdir(exist_ok=True)

        p11lib = "/usr/lib64/pkcs11/libsofthsm2.so"
        # Initialize SoftHSM2 token. An error would be raised if token with same
        # label would be created.
        cmd = ["softhsm2-util", "--init-token", "--free", "--label",
               self.name, "--so-pin", "12345678",
               "--pin", self.pin]
        run(cmd, env={"SOFTHSM2_CONF": self._softhsm2_conf}, check=True)
        logger.debug(
            f"SoftHSM token is initialized with label '{self.cardholder}'")

        # Initialize NSS db
        self._nssdb = self.card_dir.joinpath("db")
        self._nssdb.mkdir(exist_ok=True)
        run(f"modutil -create -dbdir sql:{self._nssdb} -force", check=True)
        logger.debug(f"NSS database is initialized in {self._nssdb}")

        out = run(f"modutil -list -dbdir sql:{self._nssdb}")
        if "library name: p11-kit-proxy.so" not in out.stdout:
            run(["modutil", "-force", "-add", 'SoftHSM PKCS#11', "-dbdir",
                 f"sql:{self._nssdb}", "-libfile", p11lib])
            logger.debug("SoftHSM support is added to NSS database")

        # Create systemd service
        with self._template.open() as tmp:
            with self._service_location.open("w") as f:
                f.write(tmp.read().format(username=self.cardholder,
                                          softhsm2_conf=self._softhsm2_conf,
                                          card_dir=self.card_dir))

        logger.debug(f"Service is created in {self._service_location}")
        run("systemctl daemon-reload")

        return self

    def delete(self):
        """
        Deletes the virtual card directory which contains certs, SoftHSM2 token
        and NSS database. Also removes the systemd service for virtual smart
        card.
        """
        shutil.rmtree(self.card_dir)
        logger.info(f"Virtual card dir of {self.name} removed")

        self._service_location.unlink()
        run("systemctl daemon-reload", sleep=3)
        logger.debug(f"Service {self._service_name} was removed")

        if self.dump_file.exists():
            self.dump_file.unlink()
            logger.debug(f"Removed {self.dump_file} dump file")

    def gen_csr(self):
        """
        Method for generating user specific CSR file that would be sent to the
        CA for generating the certificate. CSR is generated using 'openssl`
        command based on template CNF file.
        """
        csr_path = self.card_dir.joinpath(f"csr-{self.cardholder}.csr")
        if self.user.user_type == UserType.local:
            cmd = ["openssl", "req", "-new", "-nodes", "-key", self.key,
                   "-reqexts", "req_exts", "-config", self.cnf,
                   "-out", csr_path]
        else:
            if not self.key:
                raise SCAutolibException("Can't generate CSR because private "
                                         "key is not set")
            cmd = ["openssl", "req", "-new", "-days", "365",
                   "-nodes", "-key", self.key, "-out",
                   csr_path, "-subj", f"/CN={self.cardholder}"]
        run(cmd)
        return csr_path


class PhysicalCard(Card):
    """
    :TODO PhysicalCard is not yet tested, it's Work In Progress
        This class provides methods allowing to manipulate physical cards
        connected via removinator.
    """
    _inserted: bool = False

    name: str = None
    pin: str = None
    cardholder: str = None
    CN: str = None
    UID: str = None
    expires: str = None
    card_type: str = None
    ca_name: str = None
    slot: str = None
    uri: str = None
    card_dir: Path = None

    def __init__(self, card_data: dict = None, card_dir: Path = None):
        """
        TODO this is not yet tested, insert and remove methods need to be
            implemented with removinator
        Initialise object for physical smart card. Constructor of the base class
        is also used.
        """
        self.card_data = card_data
        # Not sure we will need following, let's see later
        self.name = card_data["name"]
        self.pin = card_data["pin"]
        self.cardholder = card_data["cardholder"]
        self.CN = card_data["CN"]
        self.UID = card_data["UID"]
#        self.expires = card_data["expires"]
#        self.card_type = card_data["card_type"]
#        self.ca_name = card_data["ca_name"]
        self.slot = card_data["slot"]
        self.uri = card_data["uri"]
        self.card_dir = card_dir
        if not self.card_dir.exists():
            raise FileNotFoundError("Card root directory doesn't exists")

        self.dump_file = LIB_DUMP_CARDS.joinpath(f"{self.name}.json")

    def __call__(self, insert: bool):
        """
        Call method for physical smart card. It would be used in the context
        manager.

        :param insert: True if the card should be inserted, False otherwise
        :type insert: bool
        """
        if insert:
            self.insert()
        return self.__enter__()

    def __enter__(self):
        """
        Start of context manager for physical smart card. The card would be
        inserted if ``insert`` parameter in constructor is specified.

        :return: self
        """
        return self

    def __exit__(self, exp_type, exp_value, exp_traceback):
        """
        End of context manager for physical smart card. If any exception was
        raised in the current context, it would be logged as an error.

        :param exp_type: Type of the exception
        :param exp_value: Value for the exception
        :param exp_traceback: Traceback of the exception
        """
        if exp_type is not None:
            logger.error("Exception in physical smart card context")
            logger.error(format_exc())
        if self._inserted:
            self.remove()

    def to_dict(self):
        """
        Customising default property for better serialisation for storing to
        JSON format.

        :return: dictionary with all values. Path objects are typed to string.
        :rtype: dict
        """
        return self.card_data

    @property
    def user(self):
        return self.cardholder

    def insert(self):
        """
        Insert physical card using removinator
        """
        ...

    def remove(self):
        """
        Remove physical card using removinator
        """
        ...



import os
import pwd
import pytest

from SCAutolib.models.user import User
from SCAutolib.utils import dump_to_json


@pytest.mark.skipif(os.getuid() != 0, reason="Requires root privileges!")
def test_add_and_remove_local_user(local_user):
    local_user.add_user()
    assert pwd.getpwnam(local_user.username)

    local_user.delete_user()
    with pytest.raises(Exception):
        pwd.getpwnam(local_user.username)


def test_dump_and_load_user(local_user):
    dump_to_json(local_user)

    user = User.load(local_user.dump_file)

    assert user.username == local_user.username
    assert user.password == local_user.password
    assert user.user_type == local_user.user_type



"""
This module contains unit tests for SSSDConf and related functions

The unit test name reflects name of tested sssd_conf method. For example
'test_create_template_exists' is unit test for method 'create' of sssd_conf.
"""
import configparser
import pathlib
import shutil
from configparser import ConfigParser
from pathlib import Path

import pytest

from SCAutolib.models.file import SSSDConf
from conftest import FILES_DIR


def load_file_to_parser(config_file: pathlib.Path):
    """
    Instantiate parser object and update it with content of config file

    :param config_file: path to config file
    :return: parser object
    """
    parser = ConfigParser()
    parser.optionxform = str
    with config_file.open() as f1:
        parser.read_file(f1)
    return parser


def parser_object_to_dict(parser_object: configparser.ConfigParser):
    """
    Transform content of parser object to dictionary

    :param parser_object: object to be transformed
    :return: dictionary reflecting content of parser object
    """
    content = {section: dict(parser_object.items(section)) for section in
               parser_object.sections()}
    return content


def compare_conf_files(file1: pathlib.Path, file2: pathlib.Path):
    """
    Compare content of two config files ignoring empty lines and reshuffling
    of content

    :param file1:  first of compared files
    :param file2:  second of compared files
    :return: True if content of files matches, False otherwise
    """
    dict1 = parser_object_to_dict(load_file_to_parser(file1))
    dict2 = parser_object_to_dict(load_file_to_parser(file2))
    assert dict1 == dict2


# TESTS
def test_check_backups(tmpdir):
    """
    Exception is raised if backup file exists
    """
    tmpfile = Path(tmpdir, "testfile")
    tmpfile.touch()
    sssd = SSSDConf()
    sssd._backup_default = tmpfile
    sssd._backup_original = tmpfile
    with pytest.raises(FileExistsError, match=f'{tmpfile} file exists'):
        sssd.check_backups()


def test_create_content_backup_original(sssd_test_prepare):
    """
    Backup of original config file is created if the file exists
    Backup of library default conf_file is created
    """
    sssd_test = sssd_test_prepare

    # create conf
    shutil.copy2(SSSDConf._template, sssd_test._conf_file)
    sssd_test.create()

    assert sssd_test._backup_default.exists()
    compare_conf_files(SSSDConf._template, sssd_test._backup_default)


def test_create_conf_updated(sssd_test_prepare):
    """
    Content of config file is updated with values from template
    if the file already exists
    """
    # create content of config based on test config file; load content to dict
    sssd_test = sssd_test_prepare
    sssd_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")

    sssd_test.create()

    # load reference content from library file to parser object
    reference_file = Path(FILES_DIR, "sssd_conf_test_result")
    reference_content = load_file_to_parser(reference_file)

    assert parser_object_to_dict(sssd_test._default_parser) == \
           parser_object_to_dict(reference_content)


def test_set_conf_add_value(sssd_test_prepare):
    """
    Tests that new value is added to existing section
    """
    sssd_test = sssd_test_prepare
    sssd_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")
    # load some content to parser object
    with sssd_test._conf_file.open() as config:
        sssd_test._default_parser.read_file(config)
    sssd_test.set("newkey", "newvalue", "testsection")
    assert sssd_test._changes["testsection"]["newkey"] == "newvalue"


def test_set_conf_add_section(sssd_test_prepare):
    """
    Tests that section is added with key and value if the section was not
    present
    """
    sssd_test = sssd_test_prepare
    sssd_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")
    # load some content to parser object
    with sssd_test._conf_file.open() as config:
        sssd_test._default_parser.read_file(config)
    sssd_test.set("newkey", "newvalue", "newsection")
    assert sssd_test._changes["newsection"]["newkey"] == "newvalue"


def test_set_conf_overwrite_value(sssd_test_prepare):
    """
    Tests that value in is overwritten if section and key already existed
    """
    sssd_test = sssd_test_prepare
    sssd_test._conf_file = Path(FILES_DIR, "sssd_conf_test_start")
    # load some content to parser object
    with sssd_test._conf_file.open() as config:
        sssd_test._default_parser.read_file(config)
    sssd_test.set("testkey", "overwrite", "testsection")
    assert sssd_test._changes["testsection"]["testkey"] == "overwrite"


def test_save_reset_parser(sssd_test_prepare):
    """
    Test that save method saves config file and reset parser object
    """
    sssd_test = sssd_test_prepare
    sssd_test.set("testkey", "overwrite", "testsection")
    sssd_test.save()
    assert sssd_test._conf_file.exists()
    assert len(sssd_test._changes.sections()) == 0


@pytest.mark.service_restart
def test_context_manager(tmpdir, sssd_test_prepare):
    """
    Test that context manager works as expected
    """
    call_key, call_value, call_section = "testkey", "testvalue", "testsection"
    sssd_test_prepare._backup_default.touch(exist_ok=True)

    with sssd_test_prepare(key=call_key, value=call_value, section=call_section):
        parser = ConfigParser()
        with sssd_test_prepare._conf_file.open() as config:
            parser.read_file(config)
        assert parser.get(call_section, call_key) == call_value

    with sssd_test_prepare._conf_file.open() as config:
        parser = ConfigParser()
        parser.read_file(config)
        with pytest.raises(configparser.NoSectionError):
            parser.get(call_section, call_key)



"""
This module provides a function (isDistro) that helps us identify the os
of the system and configure the system accordingly.
"""

import distro
from typing import Union


def isDistro(OSes: Union[str, list], version: str = None) -> bool:
    cur_id = distro.id().lower()
    cur_name = distro.name().lower()

    if isinstance(OSes, str):
        results = (OSes in cur_id) or (OSes in cur_name)
    else:
        results = False
        for item in OSes:
            if not isinstance(item, str):
                continue
            item = item.lower()
            results = results or (item in cur_id) or (item in cur_name)

    if results is False:
        return False

    if version:
        cur_major = int(distro.major_version())
        cur_minor = int(distro.minor_version()) if distro.minor_version() else 0

        if version[0] in ('<', '=', '>'):
            if version[1] == '=':
                op = version[:2]
                version = version[2:]
            else:
                op = version[0] if version[0] != '=' else '=='
                version = version[1:]
        else:
            op = '=='

        parts = version.split('.')
        major = int(parts[0])
        minor = int(parts[1]) if len(parts) > 1 else None

        if major == cur_major and minor:
            return eval("{0} {1} {2}".format(cur_minor, op, minor))
        else:
            return eval("{0} {1} {2}".format(cur_major, op, major))

    return True



import coloredlogs
import logging
import subprocess
from pathlib import Path
import time
from schema import Schema, Use, Or, And, Optional

from SCAutolib.enums import CardType, UserType

fmt = ("%(asctime)s %(name)s:%(module)s.%(funcName)s.%(lineno)d "
       "[%(levelname)s] %(message)s")
date_fmt = "%H:%M:%S"
coloredlogs.install(level="DEBUG", fmt=fmt, datefmt=date_fmt,
                    field_styles={'levelname': {'bold': True, 'color': 'blue'},
                                  'asctime': {'color': 'green'}})
logger = logging.getLogger(__name__)
# Disable logs from imported packages
logging.getLogger("paramiko").setLevel(logging.WARNING)
logging.getLogger("invoke").setLevel(logging.WARNING)
logging.getLogger("fabric").setLevel(logging.WARNING)

DIR_PATH = Path(__file__).parent
TEMPLATES_DIR = DIR_PATH.joinpath("templates")

LIB_DIR = Path("/etc/SCAutolib")
SETUP_IPA_SERVER = LIB_DIR.joinpath("ipa-install-server.sh")
LIB_BACKUP = LIB_DIR.joinpath("backup")
LIB_DUMP = LIB_DIR.joinpath("dump")
LIB_DUMP_USERS = LIB_DUMP.joinpath("users")
LIB_DUMP_CAS = LIB_DUMP.joinpath("cas")
LIB_DUMP_CARDS = LIB_DUMP.joinpath("cards")
LIB_DUMP_CONFS = LIB_DUMP.joinpath("confs")


schema_cas = Schema(And(
    Use(dict),
    # Check that CA section contains at least one and maximum
    # two entries
    lambda l: 1 <= len(l.keys()) <= 2,  # noqa: E741
    {Optional("local_ca"): {
        Optional("dir", default=Path("/etc/SCAutolib/ca")): Use(Path)},
        Optional("ipa"): {
            'admin_passwd': Use(str),
            'root_passwd': Use(str),
            Optional('ip_addr', default=None): Use(str),
            'server_hostname': Use(str),
            'client_hostname': Use(str),
            'domain': Use(str),
            'realm': Use(str.upper)}}),
    ignore_extra_keys=True)

# Specify validation schema for all users
schema_user = Schema({'name': Use(str),
                      'passwd': Use(str),
                      'user_type': Or(UserType.local, UserType.ipa)})

# Specify validation schema for all cards
schema_card = Schema({'name': Use(str),
                      'pin': Use(str),
                      Optional('card_details', default=None): Use(str),
                      'cardholder': Use(str),
                      'CN': Use(str),
                      Optional('UID', default=None): Use(str),
                      Optional('expires', default=None): Use(str),
                      'card_type': Or(CardType.virtual, CardType.physical),
                      'ca_name': Use(str),
                      Optional('ca_cert', default=None): Use(str),
                      Optional('slot', default=None): Use(str),
                      Optional('uri', default=None): Use(str),
                      Optional('cert', default=None): Use(str),
                      Optional('key', default=None): Use(str)})


def run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True,
        print_=True, return_code: list = None, sleep: int = 0, **kwargs) \
        -> subprocess.CompletedProcess:
    """
    Wrapper for subrpocess.run function. This function explicitly set several
    parameter of original function and also provides similar thing as
    subprocess.check_output do. But with having this wrapper, functionality
    of this two functions is generalized and can be changed by setting
    corresponding parameters. If there are any specific parameter of
    subprocess.run function needed to be passed to this wrapper, you can do
    it by adding same parameters names in key=value format.

    :param sleep: time to sleep after command is executed
    :type sleep: int
    :param return_code: acceptable return codes from given commands.
        If check=True, and the return code of the cmd is not in the return_code
        list a subprocess.CalledProcessError exception would be raised.
    :type return_code: list
    :param cmd: Command to be executed
    :type cmd: list or str
    :param stdout: Redirection of stdout. Default is subprocess.PIPE
    :type stdout: None or int or IO
    :param stderr: Redirection of stderr. Default is subprocess.PIPE
    :type stderr: None or int or IO
    :param check: Specifies it return code of the command would be checked for
        0 (if return code == 0). If True and return code is not 0, then
        subprocess.CalledProcessError exception would be risen. Default is
        False.
    :type check: bool
    :param print_: Specifies it stdout and stderr should be printed to the
        terminal. Log message with stdout would have debug type and stderr
        log message would have error type. Default is True.
    :type print_: bool
    :param kwargs: Other parameters to subprocess.run function

    :exception subprocess.CalledProcessError:

    :return: Completed process from subprocess.run
    :rtype: subprocess.CompletedProcess
    """
    if return_code is None:
        return_code = [0]
    if isinstance(cmd, str):
        cmd = cmd.split(" ")
    logger.debug(f"run: {' '.join([str(i) for i in cmd])}")
    out = subprocess.run(cmd, stdout=stdout, stderr=stderr, encoding="utf-8",
                         **kwargs)
    if print_:
        if out.stdout != "":
            logger.debug(out.stdout)
        if out.stderr != "":
            logger.warning(out.stderr)

    if check:
        if out.returncode not in return_code:
            logger.error(f"Unexpected return code {out.returncode}. "
                         f"Expected: {return_code}")
            raise subprocess.CalledProcessError(out.returncode, cmd)
    time.sleep(sleep)
    return out



from SCAutolib import logger
from contextlib import contextmanager
import re


@contextmanager
def assert_log(path: str, expected_log: str):
    """Asserts, that a new line in log is created, that matches given regex.

    :param path: Path to the file, that will be checked for added logs.
    :param expected_log: Regular expression
        that has to match one of the new logs.

    When the context manager starts, the log file is skipped until the end
    to ignore any previous logs.
    Then the action inside the context manager is run.
    This action should generate some logs.
    When the context manager exits,
    newly generated logs are matched to the regular expression.
    In case none of the logs match, an exception is raised.
    """
    logger.info(f'Opening log file {path}')
    with open(path) as f:
        # Move file pointer to the end of file
        f.seek(0, 2)
        p = re.compile(expected_log)

        # Run the actual actions
        try:
            yield

        finally:
            logger.info(f'Asserting regex `{expected_log}` in {path}')
            log = ''  # Only for debugging purposes

            for line in f:
                log += line
                m = p.match(line)
                if m:
                    # found the log
                    text = m.group()
                    logger.info(f'Found matching line: {text}')
                    return

            logger.debug(log)
            raise Exception('The log was not found.')
